import itertools
import random
import sys
from itertools import combinations
from math import ceil, sqrt, cos, sin, radians, acos
import numpy as np

from pychemia import pcm_log, HAS_PYHULL
from pychemia.utils.mathematics import length_vectors, angle_vectors, wrap2_pmhalf, \
    unit_vector, rotation_matrix_around_axis_angle, angle_vector
from pychemia.utils.mathematics import matrix_from_eig, vector_set_perpendicular
from pychemia.utils.periodic import covalent_radius
from pychemia.core.composition import Composition


class Lattice:
    """
    Routines to create and manipulate the lattice
    The lattice is sufficiently general to account for periodicity in 1, 2 or 3 directions.
    However many routines are only implemented for 3 directions
    The lattice contains 1, 2 or 3 vectors
    """

    def __init__(self, cell=None, periodicity=True):
        """
        Defines the lattice, basically it contains the lattice vectors arranged as rows in the cell array.
        The Lattice class provides methods to compute the reciprocal lattice (That is another Lattice object)
        and  basic routines related to the cell


        >>> cubic = Lattice()
        >>> cubic.lengths
        array([ 1.,  1.,  1.])
        >>> cubic.angles
        array([ 90.,  90.,  90.])

        >>> ortho = Lattice([1, 2, 3])
        >>> ortho.lengths
        array([ 1.,  2.,  3.])
        >>> ortho.angles
        array([ 90.,  90.,  90.])

        >>> bcc = Lattice([[0.5, 0.5, -0.5], [-0.5, 0.5, 0.5], [0.5, -0.5, 0.5]])
        >>> bcc.angles
        array([ 109.47122063,  109.47122063,  109.47122063])
        >>> bcc.lengths
        array([ 0.8660254,  0.8660254,  0.8660254])

        >>> fcc = Lattice([[0.5, 0.5, 0], [0, 0.5, 0.5], [0.5, 0, 0.5]])
        >>> fcc.lengths
        array([ 0.70710678,  0.70710678,  0.70710678])
        >>> fcc.angles
        array([ 60.,  60.,  60.])

        """
        if cell is None:
            cell = np.eye(3)
        else:
            cell = np.array(cell)
        if cell.shape == (3,):
            cell = cell * np.eye(3)
        elif cell.shape == ():
            cell = cell * np.eye(3)

        self._periodicity = None
        self.set_periodicity(periodicity)
        self._dims = sum(self._periodicity)
        assert (np.prod(np.array(cell).shape) == self.periodic_dimensions ** 2)
        self._cell = np.array(cell).reshape((self.periodic_dimensions, self.periodic_dimensions))
        self._lengths = length_vectors(self._cell)
        self._angles = angle_vectors(self._cell, units='deg')
        self._metric = None
        self._inverse = None
        self._limits_for_distance2 = None

    def __str__(self):
        ret = 'Cell='
        for i in range(3):
            for j in range(3):
                ret += "%12.3f " % (self.cell[i, j])
            ret += '\n'
            if i < 2:
                ret += 5 * ' '
        ret += '\n'
        ret += 'Angles: alpha = %12.3f\n' % self.alpha
        ret += '         beta = %12.3f\n' % self.beta
        ret += '        gamma = %12.3f\n' % self.gamma
        ret += '\n'
        ret += 'Lengths:    a = %12.3f\n' % self.a
        ret += '            b = %12.3f\n' % self.b
        ret += '            c = %12.3f\n' % self.c
        return ret

    def cartesian2reduced(self, x):
        return np.dot(x, self.inverse)

    def copy(self):
        """
        Return a copy of the object
        """
        return self.__class__(self._cell, self._periodicity)

    def distance2(self, x1, x2, option='reduced', radius=20, limits=None):

        # Compute the vector from x1 to x2
        dv = np.array(x2) - np.array(x1)

        # If we are not in reduced coordinates,
        # Convert into them
        if option != 'reduced':
            dred = self.cartesian2reduced(dv)
        else:
            dred = dv

        dwrap = wrap2_pmhalf(dred)

        if limits is None:
            limits = np.zeros(3, dtype=int)
            corners = self.get_wigner_seitz_container()

            limits[0] = min(int(ceil(max(1e-14 + abs(np.array([corners[x][0] for x in corners]))))), 5)
            limits[1] = min(int(ceil(max(1e-14 + abs(np.array([corners[x][1] for x in corners]))))), 5)
            limits[2] = min(int(ceil(max(1e-14 + abs(np.array([corners[x][2] for x in corners]))))), 5)

        ret = {}
        for i0 in np.arange(-limits[0], limits[0] + 1):
            for i1 in np.arange(-limits[1], limits[1] + 1):
                for i2 in np.arange(-limits[2], limits[2] + 1):
                    dtot = dwrap + np.array([i0, i1, i2])
                    norm2 = np.dot(np.dot(dtot, self.metric), dtot)
                    if norm2 < radius * radius:
                        ret[(i0, i1, i2)] = {'distance': sqrt(norm2), 'image': dtot}

        return ret

    @staticmethod
    def from_parameters_to_cell(a, b, c, alpha, beta, gamma):
        """
        Create a Lattice using parameters a, b and c as box lengths
        and corresponging angles alpha, beta, gamma

        :param a: (float): *a* lattice length.
        :param b: (float): *b* lattice length.
        :param c: (float): *c* lattice length.
        :param alpha: (float): *alpha* angle in degrees.
        :param beta: (float): *beta* angle in degrees.
        :param gamma: (float): *gamma* angle in degrees.

        :rtype : (Lattice) object
        """

        alpha_radians = radians(alpha)
        beta_radians = radians(beta)
        gamma_radians = radians(gamma)
        val = (cos(alpha_radians) * cos(beta_radians) - cos(gamma_radians)) / (sin(alpha_radians) * sin(beta_radians))
        # Sometimes rounding errors result in values slightly > 1.
        val = val if abs(val) <= 1 else val / abs(val)
        gamma_star = acos(val)
        cell = np.zeros((3, 3))
        cell[0] = [a * sin(beta_radians), 0.0, a * cos(beta_radians)]
        cell[1] = [-b * sin(alpha_radians) * cos(gamma_star),
                   b * sin(alpha_radians) * sin(gamma_star),
                   b * cos(alpha_radians)]
        cell[2] = [0.0, 0.0, float(c)]
        return Lattice(cell)

    def get_brillouin(self):
        return self.reciprocal().get_wigner_seitz()

    def get_path(self):

        assert (self.periodic_dimensions == 3)

        zero3 = np.zeros(3)
        x = self.cell[0, :]
        y = self.cell[1, :]
        z = self.cell[2, :]

        frame = np.array([zero3, x, x + y, y, zero3, z, x + z, x + y + z, y + z, z])

        line1 = np.array([x, x + z])
        line2 = np.array([x + y, x + y + z])
        line3 = np.array([y, y + z])

        return frame, line1, line2, line3

    def get_wigner_seitz(self):
        if HAS_PYHULL:
            from pyhull.voronoi import VoronoiTess
            points = []
            for i, j, k in itertools.product((-1, 0, 1), repeat=3):
                points.append(i * self.cell[0] + j * self.cell[1] + k * self.cell[2])
            tess = VoronoiTess(points)
            ret = []
            for r in tess.ridges:
                if r[0] == 13 or r[1] == 13:
                    ret.append([tess.vertices[i] for i in tess.ridges[r]])
            return ret
        else:
            raise NotImplementedError

    def get_wigner_seitz_container(self):
        """
        Compute the corners of the box that contains the Wigner-Seitz cell

        :return: dict : dictionary with values numpy arrays
        """
        ret = {}
        for i in itertools.product((-1, 1), repeat=3):
            ret[i] = np.dot(self.reciprocal().metric, i * np.diagonal(self.metric))
        return ret

    def minimal_distance(self, x1, x2, option='reduced'):
        distances_dict = self.distance2(x1, x2, option=option)
        mindist = sys.float_info.max
        for k in distances_dict:
            if 0 < distances_dict[k]['distance'] < mindist:
                mindist = distances_dict[k]['distance']
        return mindist

    def minimal_distances(self, reduced1, reduced2):
        """
        Computes a matrix with the minimal distances between
        two sets of points represented as reciprocal coordinates

        :param reduced1: List or array of reduced coordinates for the first
                            set of points
        :param reduced2: Second set of points

        Example:
        >>> import numpy as np
        >>> lattice = Lattice.random_cell('C2')
        >>> r1 = np.random.rand(4, 3)
        >>> r2 = np.random.rand(4, 3)
        >>> dist, close_imgs = lattice.minimal_distances(r1, r2)

        >>> solution = np.zeros((len(r1), len(r2)))

        >>> for i in range(len(r1)):
        ...        for j in range(len(r2)):
        ...                reduced1 = r1[i]
        ...                reduced2 = r2[j]+close_imgs[i, j]
        ...                cartesian1 = lattice.reduced2cartesian(reduced1)
        ...                cartesian2 = lattice.reduced2cartesian(reduced2)
        ...                diff_vector = cartesian2 - cartesian1
        ...                solution[i, j] = np.linalg.norm(diff_vector)

        >>> solution - dist < 1E-5
        array([[ True,  True,  True,  True],
               [ True,  True,  True,  True],
               [ True,  True,  True,  True],
               [ True,  True,  True,  True]], dtype=bool)

        """
        # Just in case of one single coordinate
        reduced1, reduced2 = np.atleast_2d(reduced1, reduced2)

        images = np.array([list(i) for i in itertools.product([-1, 0, 1], repeat=3)])

        red2_images = reduced2[:, None, :] + images[None, :, :]

        cartesian1 = self.reduced2cartesian(reduced1)
        cartesian2 = self.reduced2cartesian(red2_images)

        diff_vectors = cartesian2[None, :, :, :] - cartesian1[:, None, None, :]

        distances = np.sum(diff_vectors * diff_vectors, axis=3)

        close_images = np.zeros((len(reduced1), len(reduced2), 3))
        for i in range(len(reduced1)):
            for j in range(len(reduced2)):
                dij = distances[i, j]
                close_images[i, j] = images[dij == min(dij)][0]

        return np.min(distances, axis=2) ** 0.5, close_images

    def distances_in_sphere(self, x1, x2, radius, option='reduced', exclude_out_sphere=True, sort_by_distance=True):
        """
        Returns all the distances between two positions x1 and x2
        taking into account the periodicity of the cell

        :param sort_by_distance:
        :param exclude_out_sphere:
        :param x1:
        :param x2:
        :param radius:
        :param option:
        :return:
        """
        # Compute the vector from x1 to x2
        dv = np.array(x2) - np.array(x1)

        # If the positions are not in reduced coordinates,convert into them
        if option != 'reduced':
            dred = self.cartesian2reduced(dv)
        else:
            dred = dv

        dwrap = wrap2_pmhalf(dred)
        # log.debug('The wrap vector is: %7.3f %7.3f %7.3f' % tuple(dwrap))

        # We need to compute the distances between equivalent faces
        # For that we to compute the perpendicular distance between
        # two faces, the reciprocal lattice produces the inverse of
        # those distances
        recp_len = np.array(self.reciprocal().lengths)
        # The reciprocal (2*pi) is not necessary
        limits = np.ceil(radius * recp_len).astype(int)
        # log.debug('The limits are: %d %d %d' % tuple(limits))

        ret = {}
        # The total size is the product of the number in the 3 directions
        # that is double the limits plus 1 to include the zero
        total_size = np.prod(2 * limits + 1)
        # log.debug('The total size is: %d' % total_size)

        ret['distance'] = np.zeros(total_size)
        ret['image'] = np.zeros((total_size, 3), dtype=int)
        ret['dwrap'] = dwrap
        ret['limits'] = limits

        index = 0
        for i0 in np.arange(-limits[0], limits[0] + 1):
            for i1 in np.arange(-limits[1], limits[1] + 1):
                for i2 in np.arange(-limits[2], limits[2] + 1):
                    dtot = dwrap + np.array([i0, i1, i2])
                    norm2 = np.dot(np.dot(dtot, self.metric), dtot)
                    ret['distance'][index] = sqrt(norm2)
                    ret['image'][index] = np.array([i0, i1, i2])
                    index += 1

        # Exclude distances out of sphere
        if exclude_out_sphere:
            inside_sphere = ret['distance'] <= radius
            if sum(inside_sphere) > 0:
                ret['distance'] = ret['distance'][inside_sphere]
                ret['image'] = ret['image'][inside_sphere]

        if sort_by_distance:
            sorted_indices = np.argsort(ret['distance'])
            ret['distance'] = ret['distance'][sorted_indices]
            ret['image'] = ret['image'][sorted_indices]

        return ret

    @staticmethod
    def random_cell(composition):
        comp = Composition(composition)
        volume = comp.covalent_volume(packing='cubes')

        random.seed()

        # make 3 random lengths
        a = (1.0 + 0.5 * random.random())
        b = (1.0 + 0.5 * random.random())
        c = (1.0 + 0.5 * random.random())

        # now we make 3 random angles
        alpha = 60.0 + 60.0 * random.random()
        beta = 60.0 + 60.0 * random.random()
        gamma = 60.0 + 60.0 * random.random()

        lattice = Lattice().from_parameters_to_cell(a, b, c, alpha, beta, gamma)

        factor = (volume / lattice.volume) ** (1 / 3.0)
        lattice = Lattice().from_parameters_to_cell(factor * a, factor * b, factor * c, alpha, beta, gamma)
        lattice.align_with_axis()
        lattice.align_with_plane()

        return lattice

    def reciprocal(self):
        """
        Return the reciprocal cell

        :rtype : Lattice
        :return:
        """
        return self.__class__(np.linalg.inv(self.cell.T))

    def reduced2cartesian(self, x):
        return np.dot(x, self.cell)

    def set_periodicity(self, periodicity):
        if isinstance(periodicity, bool):
            self._periodicity = 3 * [periodicity]
        elif np.iterable(periodicity):
            assert (len(periodicity) == 3)
            for i in periodicity:
                assert (isinstance(i, bool))
            self._periodicity = list(periodicity)
        else:
            raise ValueError("Periodicity must be a boolean or list instead of: ", periodicity)

    @property
    def cell(self):
        """
        Return the cell as a numpy array

        :return:
        """
        return self._cell

    @property
    def periodic_dimensions(self):
        """
        Return the number of periodic dimensions

        :return: int
        """
        return self._dims

    @property
    def volume(self):
        """
        Computes the volume of the cell (3D),
        area (2D) or generalized volume for
        N dimensions

        :rtype : float
        """
        return abs(np.linalg.det(self.cell))

    @property
    def metric(self):
        if self._metric is None:
            self._metric = np.dot(self.cell, self.cell.T)
        return self._metric

    @property
    def inverse(self):
        if self._inverse is None:
            self._inverse = np.linalg.inv(self.cell)
        return self._inverse

    @property
    def alpha(self):
        return self._angles[(1, 2)]

    @property
    def beta(self):
        return self._angles[(0, 2)]

    @property
    def gamma(self):
        return self._angles[(0, 1)]

    @property
    def angles(self):
        return np.round([self.alpha, self.beta, self.gamma], 14)

    @property
    def a(self):
        return self._lengths[0]

    @property
    def b(self):
        return self._lengths[1]

    @property
    def c(self):
        return self._lengths[2]

    @property
    def lengths(self):
        """
        Return the 3 lenghts of the lattice in the same order as the lattice vectors

        :rtype : list
        """
        return self._lengths

    def align_with_axis(self, axis=0, round_decimals=14):
        a = self.cell[0]
        if axis == 0:
            b = np.array([1, 0, 0])
        elif axis == 1:
            b = np.array([0, 1, 0])
        elif axis == 2:
            b = np.array([0, 0, 1])
        else:
            raise ValueError('Axis must be an integer in (0,1,2)')
        if np.linalg.norm(np.cross(a, b)) < 1E-10:
            return
        c = unit_vector(np.cross(a, b))
        av = angle_vector(a, b)
        rotation_matrix = rotation_matrix_around_axis_angle(c, av)
        self._cell = np.dot(rotation_matrix, self.cell.T).T.round(round_decimals)

    def align_with_plane(self, axis=2, round_decimals=14):
        a = self.cell[1]

        if axis == 0:
            p1 = np.array([0, 1, 0])
            p2 = np.array([0, 0, 1])
        elif axis == 1:
            p1 = np.array([0, 0, 1])
            p2 = np.array([1, 0, 0])
        elif axis == 2:
            p1 = np.array([1, 0, 0])
            p2 = np.array([0, 1, 0])
        else:
            raise ValueError('Axis must be an integer in (0,1,2)')

        if np.linalg.norm(np.cross(p1, a)) < 1E-10:
            return
        c = unit_vector(np.cross(p1, a))
        # print c
        # A vector perpendicular to the plane
        vector_plane = unit_vector(np.cross(p1, p2))
        v1_u = unit_vector(vector_plane)
        v2_u = unit_vector(c)
        proj = np.dot(v1_u, v2_u)
        # print 'Projection', proj
        av = np.arccos(proj)
        # import math
        # print 'Angle=', math.degrees(av)
        rotation_matrix = rotation_matrix_around_axis_angle(p1, -av)
        cell = np.dot(rotation_matrix, self.cell.T).T.round(round_decimals)
        # print '-->',cell[1], vector_plane
        # print 'PROJECTION', np.dot(cell[1], vector_plane)
        if np.abs(np.dot(cell[1], vector_plane)) > 1E-10:
            # print 'Failed projection', np.dot(cell[1], vector_plane)
            # print cell
            rotation_matrix = rotation_matrix_around_axis_angle(p1, av)
            cell = np.dot(rotation_matrix, self.cell.T).T.round(round_decimals)
            if np.dot(cell[1], vector_plane) > 1E-10:
                # print 'Failed projection', np.dot(cell[1], vector_plane)
                # print cell
                pass
        self._cell = cell

    def stretch(self, symbols, rpos, tolerance=1.0, extra=0.1):
        # Dummy array that always will be overwritten
        eigv = np.array([1, 0, 0])
        lattice = self.copy()
        assert len(rpos) == len(symbols)
        natom = len(rpos)
        for i, j in combinations(range(natom), 2):
            ret = lattice.distance2(rpos[i], rpos[j])
            mindist = sys.float_info.max
            for k in ret:
                if 0 < ret[k]['distance'] < mindist:
                    mindist = ret[k]['distance']
                    eigv = ret[k]['image']

            covalent_distance = sum(covalent_radius([symbols[i], symbols[j]]))
            if mindist < tolerance * covalent_distance:
                factor = (tolerance + extra) * covalent_distance / mindist
                v1, v2, v3 = vector_set_perpendicular(eigv)
                matrix_a = matrix_from_eig(v1, v2, v3, factor, 1, 1)
                lattice = Lattice(np.dot(matrix_a, lattice.cell))
        return lattice

    def scale(self, symbols, rpos, tolerance=1.0):
        lattice = self.copy()
        factor = 1.0
        for i in range(len(rpos)):
            # This is to separate each atom from its own image
            covalent_dim = 2.0 * tolerance * covalent_radius(symbols[i])

            this_factor = covalent_dim / lattice.a
            if this_factor > factor:
                factor = this_factor
            this_factor = covalent_dim / lattice.b
            if this_factor > factor:
                factor = this_factor
            this_factor = covalent_dim / lattice.c
            if this_factor > factor:
                factor = this_factor

            for j in range(i + 1, len(rpos)):
                distance = lattice.minimal_distance(rpos[i], rpos[j])
                covalent_dim = tolerance * sum(covalent_radius([symbols[i], symbols[j]]))
                this_factor = covalent_dim / distance
                if this_factor > factor:
                    factor = this_factor
        a = lattice.a
        b = lattice.b
        c = lattice.c
        alpha = lattice.alpha
        beta = lattice.beta
        gamma = lattice.gamma
        return Lattice().from_parameters_to_cell(factor * a, factor * b, factor * c, alpha, beta, gamma)



from setuptools import setup, find_packages

setup(
    name='pychemia',
    version='0.17.3.8',
    author='Guillermo Avendano-Franco',
    author_email='gufranco@mail.wvu.edu',
    packages=find_packages(),
    url='https://github.com/MaterialsDiscovery/PyChemia',
    license='LICENSE.txt',
    description='Python framework for Materials Discovery and Design',
    long_description=open('README.md').read(),
    install_requires=['numpy >= 1.12.0',
                      'scipy >= 0.18.0',
                      'future>=0.16.0',
                      'spglib>=1.9.9',
                      'pymongo>=3.4.0'],
    keywords=["electronic", "structure", "analysis", "materials", "discovery", "metaheuristics"],
)



import doctest
import unittest

from pychemia.test.doctest_2to3 import doctest_suite


def test_structure():
    """
    DocTests (pychemia.core.structure)                           :
    """
    import pychemia.core.structure
    dt = doctest.testmod(pychemia.core.structure, verbose=True)
    assert dt.failed == 0


def test_composition():
    """
    DocTests (pychemia.core.composition)                         :
    """
    import pychemia.core.composition
    suite = unittest.TestSuite()
    suite.addTest(doctest_suite(pychemia.core.composition))
    runner = unittest.TextTestRunner(verbosity=1)
    result = runner.run(suite)
    assert result.wasSuccessful()


if __name__ == "__main__":
    test_composition()



import os
import re
import time
import json
import numpy as np
from pychemia.crystal import KPoints
from pychemia import pcm_log
from pychemia.utils.serializer import generic_serializer
from ..vasp import VaspJob
from ..outcar import read_vasp_stdout
from ...tasks import Task

__author__ = 'Guillermo Avendano-Franco'


class ElasticModuli(Task):
    def __init__(self, structure, workdir='.', binary='vasp', encut=1.3, kpoints=None, kp_density=1E4):

        self.encut = encut
        if kpoints is None:
            kp = KPoints.optimized_grid(structure.lattice, kp_density=kp_density, force_odd=True)
            self.kpoints = kp
        else:
            self.kpoints = kpoints
        self.task_params = {'encut': self.encut, 'kpoints': self.kpoints.to_dict}
        Task.__init__(self, structure=structure, task_params=self.task_params, workdir=workdir, binary=binary)

    def run(self, nparal=4):

        vj = VaspJob()
        vj.initialize(self.structure, self.workdir, self.kpoints, binary=self.binary)
        vj.clean()
        vj.job_static()
        vj.input_variables.set_density_for_restart()
        vj.input_variables.set_encut(ENCUT=self.encut, POTCAR=self.workdir + os.sep + 'POTCAR')
        vj.input_variables.variables['IBRION'] = 6
        vj.input_variables.variables['ISIF'] = 3
        vj.set_inputs()
        self.encut = vj.input_variables.variables['ENCUT']
        vj.run(use_mpi=True, mpi_num_procs=nparal)
        pcm_log.debug('Starting VASP')
        while True:
            energy_str = ''
            filename = self.workdir + os.sep + 'vasp_stdout.log'
            if os.path.exists(filename):
                vasp_stdout = read_vasp_stdout(filename=filename)
                if len(vasp_stdout['data']) > 2:
                    scf_energies = [i[2] for i in vasp_stdout['data']]
                    energy_str = ' %7.3f' % scf_energies[1]
                    for i in range(1, len(scf_energies)):
                        if scf_energies[i] < scf_energies[i - 1]:
                            energy_str += ' >'
                        else:
                            energy_str += ' <'
                    pcm_log.debug(energy_str)

            if vj.runner is not None and vj.runner.poll() is not None:
                filename = self.workdir + os.sep + 'vasp_stdout.log'
                if os.path.exists(filename):
                    vasp_stdout = read_vasp_stdout(filename=filename)
                    if len(vasp_stdout['data']) > 2:
                        scf_energies = [i[2] for i in vasp_stdout['data']]
                        energy_str += ' %7.3f' % scf_energies[-1]
                        pcm_log.debug(energy_str)
                pcm_log.debug('Execution complete')
                break
            time.sleep(5)
        vj.get_outputs()

        self.output = self.get_elastic_moduli()
        if vj.outcar.is_finished:
            self.finished = True

    def plot(self, filedir=None, file_format='pdf'):
        if filedir is None:
            filedir = self.workdir
        import matplotlib.pyplot as plt
        plt.switch_backend('agg')

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(9, 5))
        plt.subplots_adjust(left=0.04, bottom=0.03, right=0.98, top=0.98, wspace=0.1, hspace=0.1)

        data = np.array(self.output['total_elastic_moduli'])
        ax1.imshow(data, interpolation='none', cmap=plt.get_cmap('Greys'))

        for i in range(len(data)):
            for j in range(len(data[i])):
                if data[i, j] > 1E-10:
                    if data[i, j] > 0.75 * np.max(data.flatten()):
                        color = 'white'
                    else:
                        color = 'black'
                    ax1.text(i, j, int(data[i, j]), ha='center', va='center', color=color)

        ax1.set_xticks(np.arange(6))
        ax1.set_xticklabels(['XX', 'YY', 'ZZ', 'XY', 'YZ', 'ZX'])
        ax1.set_yticks(np.arange(6))
        ax1.set_yticklabels(['XX', 'YY', 'ZZ', 'XY', 'YZ', 'ZX'])
        ax1.set_xlabel('Total Elastic Moduli')
        ax1.xaxis.tick_top()

        data = np.array(self.output['elastic_moduli_contributions_from_ionic_relaxation'])
        maxdata = np.abs(np.max(data.flatten()))
        ax2.imshow(data, interpolation='none', cmap=plt.get_cmap('seismic'), vmin=-maxdata, vmax=maxdata)

        for i in range(len(data)):
            for j in range(len(data[i])):
                if np.abs(data[i, j]) > 1E-10:
                    if np.abs(data[i, j]) > 0.75 * maxdata:
                        color = 'white'
                    else:
                        color = 'black'
                    ax2.text(i, j, int(data[i, j]), ha='center', va='center', color=color)

        ax2.set_xticks(np.arange(6))
        ax2.set_xticklabels(['XX', 'YY', 'ZZ', 'XY', 'YZ', 'ZX'])
        ax2.set_yticks(np.arange(6))
        ax2.set_yticklabels(['XX', 'YY', 'ZZ', 'XY', 'YZ', 'ZX'])
        ax2.set_xlabel('Elastic Moduli contributions from Ion relax')
        ax2.xaxis.tick_top()

        plt.savefig(filedir + os.sep + 'elastic.' + file_format)
        return plt.gcf()

    def load(self, filename=None):
        if filename is None:
            filename = self.workdir + os.sep + 'task.json'
        rf = open(filename)
        data = json.load(rf)
        rf.close()
        self.task_params = data['task_params']
        self.output = data['output']
        self.encut = self.task_params['encut']
        self.kpoints = KPoints.from_dict(self.task_params['kpoints'])

    def report(self, file_format='html'):
        from lxml.builder import ElementMaker, E
        self.plot(filedir=self.workdir + os.sep + 'REPORT', file_format='jpg')

        style = """
table {
    width:70%;
}
table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
}
th, td {
    padding: 5px;
    text-align: left;
}
table#t01 tr:nth-child(even) {
    background-color: #eee;
}
table#t01 tr:nth-child(odd) {
   background-color:#fff;
}
table#t01 th	{
    background-color: black;
    color: white;
}
"""
        mech = self.get_mechanical_properties()
        table = []
        for i in mech:
            table.append(E.tr(E.td(i), E.td(mech[i]['units']),
                              E.td('%7.3f' % mech[i]['Voigt']),
                              E.td('%7.3f' % mech[i]['Reuss']),
                              E.td('%7.3f' % (0.5 * (mech[i]['Voigt'] + mech[i]['Reuss'])))))

        element_maker = ElementMaker(namespace=None, nsmap={None: "http://www.w3.org/1999/xhtml"})
        html = element_maker.html(E.head(E.title("VASP Elastic Moduli"), E.style(style)),
                                  E.body(E.h1("VASP Elastic Moduli"),
                                         E.h2('Structure'),
                                         E.pre(str(self.structure)),
                                         E.h2('Elastic Moduli'),
                                         E.p(E.img(src='elastic.jpg', width="900", height="500",
                                                   alt="Static Calculation")),
                                         E.table(E.tr(E.th('Property'), E.th('Units'), E.th('Voigt'), E.th('Reuss'),
                                                      E.th('Average')), E.tr(*tuple(table)), id="t01"), E.br()
                                         ))

        return self.report_end(html, file_format)

    def get_elastic_moduli(self):

        if not os.path.isfile(self.workdir + os.sep + 'OUTCAR'):
            print('OUTCAR file not found')
            return

        return elastic_moduli()

    def get_mechanical_properties(self):
        """
        Return the mechanical properties as a dictionary (Total Elastic Moduli)

        :rtype : dict
        """
        return mechanical_properties(self.get_elastic_moduli()['total_elastic_moduli'])


def elastic_moduli(filename='OUTCAR'):
    ret = {}
    rf = open(filename)
    data = rf.read()

    subblock = re.findall('ELASTIC MODULI CONTR FROM IONIC RELAXATION \(kBar\)[\s\d\w]*-*([\s\d\w.-]*)\n\n\n',
                          data)
    ret['elastic_moduli_contr'] = np.array(np.array(subblock[0].split()[:42]).reshape(6, -1)[:, 1:], dtype=float)

    subblock = re.findall('TOTAL ELASTIC MODULI \(kBar\)[\s\d\w]*-*([\s\d\w.-]*)\n\n\n', data)
    ret['total_elastic_moduli'] = np.array(np.array(subblock[0].split()[:42]).reshape(6, -1)[:, 1:], dtype=float)

    return generic_serializer(ret)


def mechanical_properties(elastic_moduli):
    elastic_moduli = np.array(elastic_moduli)

    # Forcing the matrix to be symmetric
    elastic_moduli = 0.5 * (elastic_moduli + np.transpose(elastic_moduli))

    elastic_moduli_inv = np.linalg.inv(elastic_moduli)

    # Voigt
    em = elastic_moduli[:3, :3]
    Kv = np.trace(em) / 90.0 + 2 * np.sum(np.triu(em, 1)) / 90.0
    Gv = np.trace(em) - np.sum(np.triu(em, 1)) + 3 * np.trace(elastic_moduli[3:6, 3:6])
    Gv /= 150

    Ev = 1.0 / 3.0 / Gv + 1.0 / 9.0 / Kv
    Ev = 1.0 / Ev

    vv = 3.0 * Gv / (3.0 * Kv + Gv)
    vv = 1.0 - vv
    vv /= 2

    # Reuss
    Kr = np.trace(elastic_moduli_inv[:3, :3]) + 2 * np.sum(np.triu(elastic_moduli_inv[:3, :3], 1))
    Kr = 0.1 / Kr

    eminv = elastic_moduli_inv[:3, :3]
    Gr = 4 * np.trace(eminv) - 4 * np.sum(np.triu(eminv, 1)) + 3 * np.trace(elastic_moduli_inv[3:6, 3:6])
    Gr = 1.5 / Gr

    Er = 1.0 / 3.0 / Gr + 1.0 / 9.0 / Kr
    Er = 1.0 / Er

    vr = 3.0 * Gr / (3.0 * Kr + Gr)
    vr = 1.0 - vr
    vr /= 2

    ret = {'Bulk modulus': {'units': 'GPa', 'Voigt': Kv, 'Reuss': Kr},
           'Shear modulus': {'units': 'GPa', 'Voigt': Gv, 'Reuss': Gr},
           'Young modulus': {'units': 'GPa', 'Voigt': Ev, 'Reuss': Er},
           'Poisson ratio': {'units': 'GPa', 'Voigt': vv, 'Reuss': vr}}

    return ret



from pychemia.utils.periodic import *


def test_utils():
    """
    Test (pychemia.utils.periodic)                               :
    """
    assert covalent_radius(1) == 0.31
    assert covalent_radius([1.1, 2]) == [0.31, 0.28]
    assert len(covalent_radius()) == 103
    assert atomic_symbol(1) == 'H'
    assert atomic_symbol([1, 2.1]) == ['H', 'He']
    assert len(atomic_symbol()) == 118


def test_atomicnumber():
    """
    Test (pychemia.utils.periodic) [atomic_number]               :
    """
    assert (atomic_number('H') == 1)
    assert (atomic_number(['H', 'He']) == [1, 2])


def test_mass():
    """
    Test (pychemia.utils.periodic) [mass]                        :
    """
    assert (mass(['H', 'He']) == [1.00794, 4.002602])
    assert (mass([1, 2]) == [1.00794, 4.002602])
    assert (mass(1) == 1.00794)


def test_symbol():
    """
    Test (pychemia.utils.periodic) [atomic_symbol]               :
    """
    assert (atomic_symbol(1) == 'H')
    assert (atomic_symbol([1, 2]) == ['H', 'He'])


def test_covalent_radius():
    """
    Test (pychemia.utils.periodic) [covalent_radius]             :
    """
    assert (covalent_radius('H') == 0.31)
    assert (covalent_radius([1, 2]) == [0.31, 0.28])
    assert (covalent_radius(['H', 'He']) == [0.31, 0.28])


def test_valence():
    """
    Test (pychemia.utils.periodic) [valence]                     :
    """
    assert (valence(1) == 1)
    assert (valence([1, 2]) == [1, 0])



#!/usr/bin/env python

import unittest
import numpy as np
import logging
from pychemia import pcm_log
from pychemia.searcher import FireFly, HarmonySearch, ParticleSwarm, GeneticAlgorithm
from pychemia.population import RealFunction
from pychemia.utils.metaheuristics import Sphere

logging.basicConfig(level=logging.DEBUG)


class SearcherTest(unittest.TestCase):

    def test_harmony(self):
        """
        Tests (pychemia.searcher.harmony)                            :
        """
        pcm_log.debug('HarmonySearch')
        mini = Sphere().minimum(3)
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=False)
        searcher = HarmonySearch(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=True)

        searcher = HarmonySearch(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        assert np.linalg.norm(np.array(searcher.population.db[searcher.population.best_candidate]['x']) - mini) < 0.2

    def test_swarm(self):
        """
        Tests (pychemia.searcher.swarm)                              :
        """
        pcm_log.debug('ParticleSwarm')
        mini = Sphere().minimum(3)
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=False)
        searcher = ParticleSwarm(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=True)
        searcher = ParticleSwarm(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        assert np.linalg.norm(np.array(searcher.population.db[searcher.population.best_candidate]['x']) - mini) < 0.2

    def test_firefly(self):
        """
        Tests (pychemia.searcher.firefly)                            :
        """
        pcm_log.debug('FireFly')
        mini = Sphere().minimum(3)
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=False)
        searcher = FireFly(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=True)
        searcher = FireFly(popu, {'delta': 0.1, 'gamma': 0.1, 'beta0': 0.8, 'alpha0': 0, 'multi_move': True},
                           generation_size=16, stabilization_limit=5)
        searcher.run()
        assert np.linalg.norm(np.array(searcher.population.db[searcher.population.best_candidate]['x']) - mini) < 0.2

    def test_genetic(self):
        """
        Tests (pychemia.searcher.genetic)                            :
        """
        pcm_log.debug('GeneticAlgorithm')
        mini = Sphere().minimum(3)
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=False)
        searcher = GeneticAlgorithm(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        popu = RealFunction(Sphere.function, 3, [-1, 1], local_minimization=True)
        searcher = GeneticAlgorithm(popu, generation_size=16, stabilization_limit=5)
        searcher.run()
        assert np.linalg.norm(np.array(searcher.population.db[searcher.population.best_candidate]['x']) - mini) < 0.2



import unittest
import pychemia
from pychemia.utils.serializer import generic_serializer
import numpy as np
import tempfile
import shutil


class PopulationTest(unittest.TestCase):
    def test_orbital(self):
        """
        Tests (pychemia.population.OrbitalDFTU)                      :
        """
        if not pychemia.db.has_connection():
            return

        pychemia_path = pychemia.__path__[0]
        abiinput = pychemia.code.abinit.InputVariables(pychemia_path + '/test/data/abinit_dmatpawu/abinit.in')
        dmatpawu = np.array(abiinput['dmatpawu']).reshape(-1, 5, 5)
        params = pychemia.population.orbitaldftu.dmatpawu2params(dmatpawu, 5)
        dmatpawu_new = pychemia.population.orbitaldftu.params2dmatpawu(params)

        self.assertLess(np.min(dmatpawu-dmatpawu_new), 0.01)

        with self.assertRaises(ValueError) as context:
            pychemia.population.orbitaldftu.OrbitalDFTU('test', '/tmp/no_abinit.in')

        with self.assertRaises(ValueError) as context:
            pychemia.population.orbitaldftu.OrbitalDFTU('test', input_path=pychemia_path +
                                                        '/test/data/abinit_01/abinit.in')

        popu = pychemia.population.orbitaldftu.OrbitalDFTU('test', input_path=pychemia_path +
                                                           '/test/data/abinit_dmatpawu/abinit.in')
        popu.pcdb.clean()

        ea = [[-0.1, -0.2, -0.3, -0.4, -0.5, -0.6, -0.7, -0.8, -0.9, -1.0],
              [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
        params['euler_angles'] = ea
        params = generic_serializer(params)
        entry_id = popu.new_entry(params)

        ea = [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
              [-0.1, -0.2, -0.3, -0.4, -0.5, -0.6, -0.7, -0.8, -0.9, -1.0],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
        params['euler_angles'] = ea
        params = generic_serializer(params)
        entry_jd = popu.new_entry(params)

        popu.add_random()
        popu.random_population(16)
        popu.cross([entry_id, entry_jd])
        print(popu)

        self.assertFalse(popu.is_evaluated(entry_id))

        popu.check_duplicates(popu.members)

        entry_idm = popu.move_random(entry_id)
        popu.get_entry(entry_idm, {'properties': 1})

        entry_imj = popu.move(entry_id, entry_jd)
        popu.get_entry(entry_imj, {'properties': 1})

        # pd = popu.to_dict
        # popu.from_dict(pd)

        tmpdir = tempfile.mkdtemp()

        for i in popu.members:
            popu.prepare_folder(i, workdir=tmpdir, source_dir=pychemia_path + '/test/data/abinit_dmatpawu')

        popu.pcdb.clean()
        shutil.rmtree(tmpdir)



import math
import numpy as np

from .searcher import Searcher
from pychemia import pcm_log


class FireFly(Searcher):
    def __init__(self, population, params=None, generation_size=32,
                 stabilization_limit=10, target_value=None):
        """
        Implementation fo the Firefly algorithm for global minimization
        This searcher uses a metric to compute the attractiveness and the vector displacement
        to move one firefly in the direction of another one

        :param population:
        :param params: (dict) Parameters to setup the Searcher. For Firefly the parameters are:
            'beta0': A value in the range (0,1) that defines how close the mobile structure will be from to
             the target structure
            'gamma': Defines the exponent of decreasing for the movement of the mobile structure
            'alpha0': Factor of scale for the random movement
            'delta': How the random change decreases with time
            'multi_move': Boolean to express if the fireflies moves following all the  other brighter ones or just
            the closest brigter firefly
        :param generation_size: (int)
        :param stabilization_limit: (int)
        :return:
        """
        # Initializing objects
        Searcher.__init__(self, population, generation_size, stabilization_limit, target_value)
        # Parameters
        self.gamma = None
        self.beta0 = None
        self.alpha0 = None
        self.delta = None
        self.multi_move = None
        self.set_params(params)

    def set_params(self, params):
        # Default parameters
        self.gamma = 0.1
        self.beta0 = 0.9
        self.alpha0 = 0.3
        self.delta = 0.9
        self.multi_move = True
        if params is None:
            params = {}
        if 'gamma' in params:
            self.gamma = params['gamma']
        if 'beta0' in params:
            self.beta0 = params['beta0']
        if 'alpha0' in params:
            self.alpha0 = params['alpha0']
        if 'delta' in params:
            self.delta = params['delta']
        if 'multi_move' in params:
            self.multi_move = params['multi_move']

    def get_params(self):
        return {'gamma': self.gamma,
                'beta0': self.beta0,
                'alpha0': self.alpha0,
                'delta': self.delta,
                'multi_move': self.multi_move}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.actives_in_generation)
        pcm_log.info(' Size of selection : %d' % len(selection))

        # For statistical purposes
        distances = []
        intensities = []
        atractiveness = []

        # Minus sign because we are searching for minima
        intensity = self.population.get_values(selection)
        for entry_id in intensity:
            intensity[entry_id] *= -1

        moves = {}
        new_selection = {}
        for entry_id in selection:
            new_selection[entry_id] = None
            moves[entry_id] = 0

        # Move all the fireflies (Except the most brightness)
        # as the selection is sorted it means that the first one will no move
        pcm_log.debug('No Moving %d %s. Intensity: %7.3f' % (0, str(selection[0]), intensity[selection[0]]))
        for i in range(1, len(selection)):
            entry_id = selection[i]
            pcm_log.debug('Moving %d %s. Intensity: %7.3f' % (i, str(entry_id), intensity[entry_id]))

            # Moving in the direction of all the brighter fireflies
            if self.multi_move:

                for j in list(range(0, i))[::-1]:
                    entry_jd = selection[j]

                    distance = self.population.distance(entry_id, entry_jd)
                    beta = self.beta0 * math.exp(-self.gamma * distance * distance)
#                    The variation of attractiveness \beta with the distance r
                    pcm_log.debug('[%s] Distance: %7.3f. Intensity: %7.3f. Atractiveness: %7.3f' % (str(entry_jd),
                                                                                                    distance,
                                                                                                    intensity[entry_jd],
                                                                                                    beta))
                    # Collecting Statistics
                    distances.append(distance)
                    intensities.append(intensity[entry_jd])
                    atractiveness.append(beta)

                    if new_selection[entry_id] is None:
                        new_selection[entry_id] = self.population.move(entry_id, entry_jd, factor=beta, in_place=False)
                        if self.alpha0 > 0:
                            factor = self.alpha0 * (self.delta ** self.current_generation)
                            self.population.move_random(new_selection[entry_id], factor=factor, in_place=True)
                    else:
                        self.population.move(new_selection[entry_id], entry_jd, factor=beta, in_place=True)
                        if self.alpha0 > 0:
                            factor = self.alpha0 * (self.delta ** self.current_generation)
                            self.population.move_random(new_selection[entry_id], factor=factor, in_place=True)
#                    print(new_selection)
                    moves[entry_id] += 1

            # Moving in the direction of the closets brighter firefly
            else:
                distances = [self.population.distance(entry_id, entry_jd) for entry_jd in selection[:i]]
                distance = min(distances)
                target = selection[distances.index(distance)]
                beta = self.beta0 * math.exp(-self.gamma * distance * distance)
                # The variation of attractiveness \beta with the distance r
                pcm_log.debug('[%s] Distance: %7.3f. Intensity: %7.3f. Atractiveness: %7.3f' %
                              (str(entry_jd), distance, intensity[entry_jd], beta))

                new_selection[entry_id] = self.population.move(entry_id, target, factor=beta, in_place=False)
                factor = self.alpha0 * (self.delta ** self.current_generation)
                self.population.move_random(new_selection[entry_id], factor=factor, in_place=True)
                moves[entry_id] += 1

        if len(distances) > 0:
            pcm_log.info('+----------------+--------------+-------------+-------------+')
            pcm_log.info('+                |    Minimum   |   Maximum   |   Average   |')
            pcm_log.info('+----------------+--------------+-------------+-------------+')
            pcm_log.info('+ Distances      |    %7.2f   |   %7.2f   |   %7.2f   |' % (np.min(distances),
                                                                                      np.max(distances),
                                                                                      np.average(distances)))
            pcm_log.info('+ Intensities    |    %7.2f   |   %7.2f   |   %7.2f   |' % (np.min(intensities),
                                                                                      np.max(intensities),
                                                                                      np.average(intensities)))
            pcm_log.info('+ Attractiveness |    %7.2f   |   %7.2f   |   %7.2f   |' % (np.min(atractiveness),
                                                                                      np.max(atractiveness),
                                                                                      np.average(atractiveness)))
            pcm_log.info('+----------------+--------------+-------------+-------------+')

        for entry_id in selection:
            if new_selection[entry_id] is not None:
                self.replace_by_other(entry_id, new_selection[entry_id],
                                      reason='Moved %d times' % moves[entry_id])
                self.population.enable(new_selection[entry_id])
            else:
                self.pass_to_new_generation(entry_id, reason='The best')
